package org.antlr.tool;

package .var.folders.cl.9twcgjfx4vsg5zl73c6wxmvm0000gn.T;

import java.io.BufferedReader;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Random;
import java.util.Stack;
import org.antlr.Tool;
import org.antlr.analysis.Label;
import org.antlr.analysis.NFAState;
import org.antlr.analysis.RuleClosureTransition;
import org.antlr.analysis.Transition;
import org.antlr.misc.IntervalSet;
import org.antlr.misc.Utils;
import org.antlr.tool.CompositeGrammar;
import org.antlr.tool.Grammar;
import org.antlr.tool.Rule;

public class RandomPhrase {
  public static final boolean debug = false;
  
  protected static Random random;
  
  protected static void randomPhrase(Grammar g, List<Integer> tokenTypes, String startRule) {
    NFAState state = g.getRuleStartState(startRule);
    NFAState stopState = g.getRuleStopState(startRule);
    Stack<NFAState> ruleInvocationStack = new Stack<NFAState>();
    while (state != stopState || !ruleInvocationStack.isEmpty()) {
      if (state.getNumberOfTransitions() == 0)
        return; 
      if (state.isAcceptState()) {
        NFAState invokingState = ruleInvocationStack.pop();
        RuleClosureTransition invokingTransition = (RuleClosureTransition)invokingState.transition[0];
        state = invokingTransition.followState;
        continue;
      } 
      if (state.getNumberOfTransitions() == 1) {
        Transition t0 = state.transition[0];
        if (t0 instanceof RuleClosureTransition) {
          ruleInvocationStack.push(state);
        } else if (t0.label.isSet() || t0.label.isAtom()) {
          tokenTypes.add(getTokenType(t0.label));
        } 
        state = (NFAState)t0.target;
        continue;
      } 
      int decisionNumber = state.getDecisionNumber();
      if (decisionNumber == 0) {
        System.out.println("weird: no decision number but a choice node");
        continue;
      } 
      int n = g.getNumberOfAltsForDecisionNFA(state);
      int randomAlt = random.nextInt(n) + 1;
      NFAState altStartState = g.getNFAStateForAltOfDecision(state, randomAlt);
      Transition t = altStartState.transition[0];
      state = (NFAState)t.target;
    } 
  }
  
  protected static Integer getTokenType(Label label) {
    if (label.isSet()) {
      IntervalSet typeSet = (IntervalSet)label.getSet();
      int randomIndex = random.nextInt(typeSet.size());
      return Integer.valueOf(typeSet.get(randomIndex));
    } 
    return Utils.integer(label.getAtom());
  }
  
  public static void main(String[] args) {
    if (args.length < 2) {
      System.err.println("usage: java org.antlr.tool.RandomPhrase grammarfile startrule");
      return;
    } 
    String grammarFileName = args[0];
    String startRule = args[1];
    long seed = System.currentTimeMillis();
    if (args.length == 3) {
      String seedStr = args[2];
      seed = Long.parseLong(seedStr);
    } 
    try {
      random = new Random(seed);
      CompositeGrammar composite = new CompositeGrammar();
      Tool tool = new Tool();
      Grammar parser = new Grammar(tool, grammarFileName, composite);
      composite.setDelegationRoot(parser);
      FileReader fr = new FileReader(grammarFileName);
      BufferedReader br = new BufferedReader(fr);
      parser.parseAndBuildAST(br);
      br.close();
      parser.composite.assignTokenTypes();
      parser.composite.defineGrammarSymbols();
      parser.composite.createNFAs();
      List<? extends Collection<? extends Rule>> leftRecursiveRules = parser.checkAllRulesForLeftRecursion();
      if (leftRecursiveRules.size() > 0)
        return; 
      if (parser.getRule(startRule) == null) {
        System.out.println("undefined start rule " + startRule);
        return;
      } 
      String lexerGrammarText = parser.getLexerGrammar();
      Grammar lexer = new Grammar(tool);
      lexer.importTokenVocabulary(parser);
      lexer.fileName = grammarFileName;
      if (lexerGrammarText != null) {
        lexer.setGrammarContent(lexerGrammarText);
      } else {
        System.err.println("no lexer grammar found in " + grammarFileName);
      } 
      lexer.buildNFA();
      leftRecursiveRules = lexer.checkAllRulesForLeftRecursion();
      if (leftRecursiveRules.size() > 0)
        return; 
      List<Integer> tokenTypes = new ArrayList<Integer>(100);
      randomPhrase(parser, tokenTypes, startRule);
      System.out.println("token types=" + tokenTypes);
      for (int i = 0; i < tokenTypes.size(); i++) {
        Integer ttypeI = tokenTypes.get(i);
        int ttype = ttypeI.intValue();
        String ttypeDisplayName = parser.getTokenDisplayName(ttype);
        if (Character.isUpperCase(ttypeDisplayName.charAt(0))) {
          List<Integer> charsInToken = new ArrayList<Integer>(10);
          randomPhrase(lexer, charsInToken, ttypeDisplayName);
          System.out.print(" ");
          for (int j = 0; j < charsInToken.size(); j++) {
            Integer cI = charsInToken.get(j);
            System.out.print((char)cI.intValue());
          } 
        } else {
          String literal = ttypeDisplayName.substring(1, ttypeDisplayName.length() - 1);
          System.out.print(" " + literal);
        } 
      } 
      System.out.println();
    } catch (Error er) {
      System.err.println("Error walking " + grammarFileName + " rule " + startRule + " seed " + seed);
      er.printStackTrace(System.err);
    } catch (Exception e) {
      System.err.println("Exception walking " + grammarFileName + " rule " + startRule + " seed " + seed);
      e.printStackTrace(System.err);
    } 
  }
}
